#include <aio.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define READERS 10
#define FILES 10
#define BUFSIZE 65536

struct reader
{
	struct aiocb aiocb;
};

/*
 * Open a random file in nonblocking mode.
 */
int
get_file()
{
	char name[64];
	int fd, flags, num;

	num = (random() % FILES) + 1;
	snprintf(name, sizeof(name), "cdn/%05d", num);

	/*
	 * Assume fd 0 is in use and won't be returned, since we use fd = 0
	 * to indicate the slot is unused.
	 */
	if ((fd = open(name, O_RDONLY)) <= 0)
		err(1, "open %s", name);
//	flags = fcntl(fd, F_GETFL, 0);
//	fcntl(fd, F_SETFL, flags | O_NONBLOCK);

	return (fd);
}

void
start_reader(struct reader *r)
{
	r->aiocb.aio_fildes = get_file();
	r->aiocb.aio_offset = 0;
	if (aio_read(&r->aiocb) < 0)
		err(1, "aio_read");
}


int
main(int main __unused, char *argv[] __unused)
{
	struct reader reader[READERS];

	memset(reader, 0, sizeof(reader));
	for (int i = 0; i < READERS; ++i) {
		struct aiocb *cb = &reader[i].aiocb;
		if (!(cb->aio_buf = malloc(BUFSIZE)))
			err(1, "malloc");
		cb->aio_nbytes = BUFSIZE;

		start_reader(&reader[i]);
	}
	int i = 0;
	int reads = 0;
	for (;;) {
		struct aiocb *cb = &reader[i].aiocb;
		if (aio_error(cb) != EINPROGRESS) {
			int ret;
			reads++;
			if ((ret = aio_return(cb)) > 0) {
				cb->aio_offset += BUFSIZE;
				if (aio_read(cb) < 0)
					err(1, "aio_read");
			} else if (errno == ENOENT) {
				close(cb->aio_fildes);
				start_reader(&reader[i]);
			} else {
				printf("i=%d offset=%u\n", i, cb->aio_offset);
				err(1, "aio_return");
			}
		}
		i = (i + 1) % READERS;
		if (reads == 10000)
			break;
	}

	return (0);
}
